import pytest
import requests


@pytest.fixture()
def stateful_handler(request):
    # same check for dynamically supplied parameters
    if hasattr(request, "param"):
        responses = request.param
    else:
        responses = ["OK", "BAD"]

    # Create a generator out of sequence of respnoses
    val_generator = (v for v in responses)

    # Handler returns next item in generator, and state is maintained until fixture is destroyed
    def my_handler(req):
        return next(val_generator)

    return my_handler


@pytest.fixture
def httpserver_dynamic(httpserver, stateful_handler):
    # https://pytest-httpserver.readthedocs.io/en/latest/howto.html#using-custom-request-handler
    httpserver.expect_request("/test", method="GET").respond_with_handler(
        stateful_handler
    )
    return httpserver


@pytest.fixture
def url_dynamic(httpserver_dynamic):
    return httpserver_dynamic.url_for("/test")


# Test that the handler steps through default generator items on successive calls
def test_handler(stateful_handler):
    assert stateful_handler(None) == "OK"
    assert stateful_handler(None) == "BAD"
    with pytest.raises(
        StopIteration
    ):  # because the handler is popping the next item from a generator, it's raising a different exception when it hits the end
        stateful_handler(None)


# Test that the handler steps through provided generator items on successive calls
@pytest.mark.parametrize(
    "stateful_handler", [["HELLO", "WORLD", "GOODBYE"]], indirect=True
)
def test_handler_parameterized(stateful_handler):
    assert stateful_handler(None) == "HELLO"
    assert stateful_handler(None) == "WORLD"
    assert stateful_handler(None) == "GOODBYE"
    with pytest.raises(
        StopIteration
    ):  # because the handler is popping the next item from a generator, it's raising a different exception when it hits the end
        stateful_handler(None)


# Test that the stateful handler works with httpserver
def test_confirm_sequence(httpserver_dynamic, url_dynamic):
    r = requests.get(url_dynamic)
    assert r.status_code == 200
    assert r.text == "OK"

    r = requests.get(url_dynamic)
    assert r.status_code == 200
    assert r.text == "BAD"

    r = requests.get(url_dynamic)
    with pytest.raises(StopIteration):
        httpserver_dynamic.check()


# Test that the stateful handler works with httpserver
@pytest.mark.parametrize("stateful_handler", [["HELLO", "WORLD"]], indirect=True)
def test_confirm_sequence_dynamic(httpserver_dynamic, url_dynamic):
    r = requests.get(url_dynamic)
    assert r.text == "HELLO"

    r = requests.get(url_dynamic)
    assert r.text == "WORLD"

    r = requests.get(url_dynamic)
    with pytest.raises(StopIteration):
        httpserver_dynamic.check()



import logging
import re
from dataclasses import dataclass, field
from typing import Any, List, Tuple

logger = logging.getLogger(__name__)


class RatesStatus:

    RATES_NAMES = {
        "TamingSpeedMultiplier": "Taming",
        "HarvestAmountMultiplier": "Harvesting",
        "XPMultiplier": "XP",
        "MatingIntervalMultiplier": "Mating Interval",
        "BabyMatureSpeedMultiplier": "Maturation",
        "EggHatchSpeedMultiplier": "Hatching",
        "BabyCuddleIntervalMultiplier": "Cuddle Interval",
        "BabyImprintAmountMultiplier": "Imprinting",
        "HexagonRewardMultiplier": "Hexagon Reward",
    }

    def __init__(
        self,
        **kwargs,
    ):
        expected, extras = self.get_expected_and_extras(kwargs)

        for k, v in expected.items():
            setattr(self, k, v)

        self.extras = extras

    def __eq__(self, __o: object) -> bool:
        return self.__dict__ == __o.__dict__

    @staticmethod
    def parse_raw(raw: str) -> dict:
        parsed_vals = {
            key.strip(): val.strip()
            for key, val in [line.split("=") for line in raw.strip().splitlines()]
        }
        return parsed_vals

    @staticmethod
    def get_expected_and_extras(parsed_vals_dict: dict) -> Tuple[dict, dict]:
        expected = {
            k: v for k, v in parsed_vals_dict.items() if k in RatesStatus.RATES_NAMES
        }
        extras = {
            k: v
            for k, v in parsed_vals_dict.items()
            if k not in RatesStatus.RATES_NAMES and k != "extras"
        }

        if "extras" in parsed_vals_dict:
            extras.update(parsed_vals_dict.get("extras"))
        logger.debug(f"Splitting rates; expected: {expected}, extras: {extras}")
        return expected, extras

    @classmethod
    def from_raw(cls, raw_txt: str):
        val_dict = cls.parse_raw(raw_txt)

        return cls.from_dict(val_dict)

    @classmethod
    def from_dict(cls, val_dict):
        expected, extras = cls.get_expected_and_extras(val_dict)

        return cls(**expected, **extras)

    def update_vals(self, vals, raw=False):

        parsed_vals = self.parse_raw(vals) if raw else vals

        expected, extras = self.get_expected_and_extras(parsed_vals)

        for k, v in expected.items():
            setattr(self, k, v)

        self.extras.update(extras)

    def to_dict(self, include_extras=False):
        expected, extras = self.get_expected_and_extras(self.__dict__)

        output_dict = expected

        if include_extras:
            output_dict.update(extras)

        return output_dict

    def to_raw(self):
        return "\n".join([f"{k}={v}" for k, v in self.to_dict().items()])

    def get_diff(self, newrates: "RatesStatus"):
        return RatesDiff.from_statuses(self, newrates)


@dataclass
class RatesDiffItem:
    key: str
    old_val: Any
    new_val: Any
    is_extra: bool = False


@dataclass
class RatesDiff:
    old: RatesStatus = field(default_factory=RatesStatus)
    new: RatesStatus = field(default_factory=RatesStatus)
    items: List[RatesDiffItem] = field(default_factory=list)

    @classmethod
    def from_statuses(cls, old, new):
        old_dict = old.to_dict()
        new_dict = new.to_dict()

        return RatesDiff(
            items=sorted(
                [
                    RatesDiffItem(
                        key=k,
                        old_val=old_dict.get(k),
                        new_val=new_dict.get(k),
                        is_extra=k not in RatesStatus.RATES_NAMES,
                    )
                    for k, _ in set(new_dict.items()) - set(old_dict.items())
                ],
                key=lambda x: x.key,
            ),
            old=old,
            new=new,
        )

    # highlights changed rates in embed
    def to_embed(self):

        rates_dict = self.old.to_dict(include_extras=False)

        # bold updated rates
        updated_rates = {item.key: f"**{item.new_val}**" for item in self.items}

        rates_dict.update(updated_rates)

        # rename keys and format embed description
        embed_description = "\n".join(
            [
                "{} × {}".format(re.sub(r"\.0", "", v), self.old.RATES_NAMES.get(k, k))
                for k, v in rates_dict.items()
            ]
        )

        return embed_description



import os
import shutil

import pytest

SAMPLE_BANSUMMARY_PATH_1 = "tests/data/bansummary-1.txt"
SAMPLE_BANSUMMARY_PATH_2 = "tests/data/bansummary-2.txt"
LAST_BANSUMMARY_PATH = "tests/data/last_bans.txt"

SAMPLE_DYNAMICCONFIG_PATH_1 = "tests/data/dynamicconfig-1.ini"
SAMPLE_DYNAMICCONFIG_PATH_2 = "tests/data/dynamicconfig-2.ini"
LAST_DYNAMICCONFIG_PATH = "tests/data/last_rates.txt"

# bans
@pytest.fixture
def sample_bansummary_1():
    with open(SAMPLE_BANSUMMARY_PATH_1) as f:
        return f.read()


@pytest.fixture
def sample_bansummary_2():
    with open(SAMPLE_BANSUMMARY_PATH_2) as f:
        return f.read()


@pytest.fixture
def last_bans():
    shutil.copy(SAMPLE_BANSUMMARY_PATH_1, LAST_BANSUMMARY_PATH)
    with open(LAST_BANSUMMARY_PATH) as f:
        yield f.read()
    os.remove(LAST_BANSUMMARY_PATH)


# rates
@pytest.fixture
def sample_dynamicconfig_1():
    with open(SAMPLE_DYNAMICCONFIG_PATH_1) as f:
        return f.read()


@pytest.fixture
def sample_dynamicconfig_2():
    with open(SAMPLE_DYNAMICCONFIG_PATH_2) as f:
        return f.read()


@pytest.fixture
def sample_application_config_yaml():
    return "tests/data/application-config-1.yaml"


@pytest.fixture
def sample_application_config_comma_yaml():
    return "tests/data/application-config-2.yaml"


@pytest.fixture
def application_config_broken():
    return "tests/data/application-config-broken.yaml"


@pytest.fixture
def sample_application_config_json():
    return "tests/data/application-config-1.json"


@pytest.fixture
def last_rates():
    shutil.copy(SAMPLE_DYNAMICCONFIG_PATH_1, LAST_DYNAMICCONFIG_PATH)
    with open(LAST_DYNAMICCONFIG_PATH) as f:
        yield f.read()
    os.remove(LAST_DYNAMICCONFIG_PATH)



import asyncio
import logging
import os

import aiohttp
import discord
from discord.ext import commands

logger = logging.getLogger(__name__)
# create cog class
class Bans(commands.Cog):
    def __init__(self, client):
        self.client = client

        self.webpage_url = client.config.bans_url
        self.channel_id = client.config.bans_channel_id
        self.polling_delay = client.config.polling_delay
        self.allowed_roles = client.config.allowed_roles
        self.data_dir = os.path.expanduser(client.config.data_dir)
        self.output_path = os.path.join(self.data_dir, "last_bans.txt")

        # Create parent directory for persistent data if it doesn't exist yet
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

    async def webpage_changed(self, response):
        # read last_bans
        if os.path.exists(self.output_path):
            with open(self.output_path) as f:
                last_bans = f.read()
        else:
            last_bans = ""
            logger.info("First run, skipping embed update")
            with open(self.output_path, "w+") as f:
                f.write(response)
            return False

        # compare responses, use splitlines to handle carriage returns and newlines
        if ("".join(response.splitlines())) == ("".join(last_bans.splitlines())):
            return False
        else:
            # update text file and embed
            with open(self.output_path, "w+") as f:
                f.write(response)
            return True

    async def send_embed(self):
        # read last_bans

        with open(self.output_path) as f:
            last_bans = f.read()
        # generate embed
        embed = discord.Embed(title="ARK Ban Summary", color=0x069420)
        embed.description = last_bans
        embed.set_image(url="https://i.stack.imgur.com/Fzh0w.png")
        # send embed
        channel = self.client.get_channel(self.channel_id)
        await channel.send(embed=embed)

    # Events
    @commands.Cog.listener()
    async def on_ready(self):
        logger.info("Cog Ready: Bans")

        while True:
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(
                        self.webpage_url,
                        headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
                    ) as response:
                        response = await response.text()
                        if await self.webpage_changed(response):
                            logger.info("Ban summary updated.")
                            await self.send_embed()
                        else:
                            logger.info("Ban summary not updated.")
            except Exception as e:
                logger.error(f"Error checking webpage: {e}")

            await asyncio.sleep(self.polling_delay)


# add cog to client
def setup(client):
    client.add_cog(Bans(client))



import discord
from discord.ext import commands
from discord.message import MessageType

ALLOWED_ROLES = ["Administrator", "Tech Administrator", "dot"]

# create cog class
class Test(commands.Cog):
    def __init__(self, client):
        self.client = client

    # Events
    @commands.Cog.listener()
    async def on_ready(self):
        print("Cog Ready: Test")

    # Commands
    @commands.has_any_role(*ALLOWED_ROLES)
    @commands.command()  # create a command within cog
    async def test(self, ctx):

        # generate embed
        embed = discord.Embed(title="Test", color=0x069420)
        embed.description = "Bot online!"

        await ctx.send(embed=embed)


# add cog to client
def setup(client):
    client.add_cog(Test(client))



import copy

import pytest

from poglink.bot import BotConfig, ConfigurableBot
from poglink.error import ConfigReadError


@pytest.fixture
def sample_config_dict():
    return {
        "token": "abcd",
        "rates_channel_id": "1234",
        "bans_channel_id": "4321",
        "polling_delay": 69,
        "allowed_roles": ["admin"],
        "rates_urls": ["www.arkrates.com"],
        "data_dir": "/my/fake/dir",
    }


@pytest.fixture
def sample_config_dict_singular(sample_config_dict):
    config = copy.deepcopy(sample_config_dict)
    config.update({"allowed_roles": "regular_user", "rates_urls": "www.google.com"})
    return config


def test_botconfig_from_dict(sample_config_dict):
    botconfig = BotConfig.from_dict(sample_config_dict)

    assert botconfig.token == "abcd"
    assert botconfig.rates_channel_id == "1234"
    assert botconfig.bans_channel_id == "4321"
    assert botconfig.polling_delay == 69
    assert botconfig.allowed_roles == ["admin"]
    assert botconfig.rates_urls == ["www.arkrates.com"]
    assert botconfig.data_dir == "/my/fake/dir"


def test_botconfig_singular_vals(sample_config_dict_singular):
    botconfig = BotConfig.from_dict(sample_config_dict_singular)

    assert botconfig.allowed_roles == ["regular_user"]
    assert botconfig.rates_urls == ["www.google.com"]


def test_botconfig_from_file(
    sample_application_config_yaml, sample_application_config_json
):
    botconfig_yaml = BotConfig.from_file(sample_application_config_yaml)
    botconfig_json = BotConfig.from_file(sample_application_config_json)

    # YAML and JSON formats both equivalent
    assert botconfig_yaml == botconfig_json

    # Expected values are present
    assert botconfig_yaml.token == "abcd"
    assert botconfig_yaml.rates_channel_id == "1234"
    assert botconfig_yaml.polling_delay == 60
    assert botconfig_yaml.allowed_roles == ["admin", "regular_users"]
    assert botconfig_yaml.rates_urls == ["http://arkdedicated.com/dynamicconfig.ini"]
    assert botconfig_yaml.data_dir == None


def test_configurable_bot(sample_config_dict):
    bot = ConfigurableBot(command_prefix=".test", config_dict=sample_config_dict)

    assert bot.config.token == "abcd"
    assert bot.config.rates_channel_id == "1234"
    assert bot.config.bans_channel_id == "4321"
    assert bot.config.polling_delay == 69
    assert bot.config.allowed_roles == ["admin"]
    assert bot.config.rates_urls == ["www.arkrates.com"]
    assert bot.config.data_dir == "/my/fake/dir"


def test_comma_separation(sample_application_config_comma_yaml):
    botconfig = BotConfig.from_file(sample_application_config_comma_yaml)

    assert botconfig.allowed_roles == ["admin", "regular_users"]
    assert botconfig.rates_urls == [
        "http://arkdedicated.com/dynamicconfig.ini",
        "http://www.google.com",
    ]


def test_config_read_error(application_config_broken):
    with pytest.raises(ConfigReadError):
        botconfig = BotConfig.from_file(application_config_broken)



import os

import discord
import discord.ext.test as dpytest
import pytest

from poglink.cogs.rates import EMBED_IMAGE, Rates
from poglink.error import RatesFetchError, RatesProcessError
from poglink.models.rates import RatesDiffItem

SAMPLE_DYNAMICCONFIG_PATH_1 = "tests/data/dynamicconfig-1.ini"
SAMPLE_DYNAMICCONFIG_PATH_2 = "tests/data/dynamicconfig-2.ini"

with open(SAMPLE_DYNAMICCONFIG_PATH_1) as f:
    SAMPLE_DYNAMICCONFIG_1 = f.read()

with open(SAMPLE_DYNAMICCONFIG_PATH_2) as f:
    SAMPLE_DYNAMICCONFIG_2 = f.read()


@pytest.fixture()
def rates_cog(sample_bot):
    rates_cog = Rates(sample_bot)
    yield rates_cog

    # Clean up
    for f in rates_cog.output_paths:
        try:
            os.remove(f)
        except:
            pass


@pytest.mark.asyncio
async def test_send_embed(rates_cog):
    # Define an embed that matches what should be published by the cog
    sample_embed = discord.Embed(
        description="test",
        title="ARK's Official server rates have just been updated!",
        color=0x63BCC3,
    )
    sample_embed.set_image(url=EMBED_IMAGE)

    # Publish the embed using the Rates cog
    await rates_cog.send_embed(
        description="test", url="www.mysite.com/dynamicconfig.ini"
    )

    # Ensure embed appears in the queue and nothing else.
    assert dpytest.verify().message().embed(embed=sample_embed)
    assert dpytest.verify().message().nothing()


@pytest.mark.parametrize(
    "sequential_handler",
    [
        (
            SAMPLE_DYNAMICCONFIG_1,
            SAMPLE_DYNAMICCONFIG_1,
            SAMPLE_DYNAMICCONFIG_1,
            SAMPLE_DYNAMICCONFIG_2,
            SAMPLE_DYNAMICCONFIG_2,
        )
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_compare_and_notify_all(rates_cog):
    sample_embed = discord.Embed(
        description="3 × Taming\n**2** × Harvesting\n3 × XP\n**0.7** × Mating Interval\n3 × Maturation\n3 × Hatching\n0.6 × Cuddle Interval\n3 × Imprinting\n**2.5** × Hexagon Reward",
        title="ARK's Official server rates have just been updated!",
        color=0x63BCC3,
    )
    sample_embed.set_image(url=EMBED_IMAGE)

    # First request; no diff. First time seeing rates
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 2nd request; no diff. Rates are now stable since they've been observed 2 times in a row
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 3rd request; still no diff. No change to stable rates
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 4th request; new rates found, but not yet stable
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 5th request; no diff from previous, which means new rates are stable. Different from previous stable rates, so embed is published
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().embed(embed=sample_embed)

    # Only one embed was published; queue is empty after consuming message above
    assert dpytest.verify().message().nothing()


@pytest.mark.parametrize(
    "sequential_handler",
    [
        (
            SAMPLE_DYNAMICCONFIG_2,
            SAMPLE_DYNAMICCONFIG_2,
            SAMPLE_DYNAMICCONFIG_1,
            SAMPLE_DYNAMICCONFIG_1,
        )
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_compare_and_notify_all_reverse(rates_cog):
    sample_embed = discord.Embed(
        description="3 × Taming\n**3** × Harvesting\n3 × XP\n**0.6** × Mating Interval\n3 × Maturation\n3 × Hatching\n0.6 × Cuddle Interval\n3 × Imprinting\n**1.5** × Hexagon Reward",
        title="ARK's Official server rates have just been updated!",
        color=0x63BCC3,
    )
    sample_embed.set_image(url=EMBED_IMAGE)

    # First request; no diff. First time seeing rates
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 2nd request; no diff. Rates are now stable since they've been observed 2 times in a row
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 3rd request; still no diff. No change to stable rates
    await rates_cog.compare_and_notify_all()
    assert dpytest.verify().message().nothing()

    # 4th request; no diff from previous, which means new rates are stable. Different from previous stable rates, so embed is published
    await rates_cog.compare_and_notify_all()
    print(rates_cog.last_rates[0].to_dict())
    assert dpytest.verify().message().embed(embed=sample_embed)

    # Only one embed was published; queue is empty after consuming message above
    assert dpytest.verify().message().nothing()


@pytest.mark.asyncio
async def test_rates_get_current_rates_bad_fetch(rates_cog):
    output_path = rates_cog.output_paths[0]

    with pytest.raises(RatesFetchError):
        rates_diff = await rates_cog.get_current_rates(
            "http://localhost:5000/bogus-url.txt"
        )


@pytest.mark.asyncio
async def test_rates_get_current_rates_bad_parse(rates_cog):
    output_path = rates_cog.output_paths[0]

    with pytest.raises(RatesProcessError):
        rates_diff = await rates_cog.get_current_rates("https://www.google.com")



import os
import tempfile

import pytest

from poglink.utils import rotate_backups


@pytest.fixture(scope="module")
def tmpdir():
    with tempfile.TemporaryDirectory() as tmp:
        yield tmp


@pytest.fixture
def main_filename(tmpdir):
    filename = os.path.join(tmpdir, "file.txt")
    return filename


def test_rotate_backup(tmpdir, main_filename):
    # Starting with empty directory
    assert len(os.listdir(tmpdir)) == 0

    # Check single file is rotated
    with open(os.path.join(tmpdir, "file.txt"), "w") as f:
        f.write("2")
    rotate_backups(main_filename, 3)
    assert ["file.txt.1"] == os.listdir(tmpdir)

    # Check two files are rotated
    with open(os.path.join(tmpdir, "file.txt"), "w") as f:
        f.write("2")
    rotate_backups(main_filename, 3)
    assert {"file.txt.2", "file.txt.1"} == set(os.listdir(tmpdir))

    # Check 3rd file is discarded
    with open(os.path.join(tmpdir, "file.txt"), "w") as f:
        f.write("3")
    rotate_backups(main_filename, 3)
    assert {"file.txt.2", "file.txt.1"} == set(os.listdir(tmpdir))

    # Check that rotation still works without base file
    rotate_backups(main_filename, 3)
    assert ["file.txt.2"] == os.listdir(tmpdir)
    with open(os.path.join(tmpdir, "file.txt.2")) as f:
        assert "3" == f.read()



import asyncio
import logging
import os
import re
import time
from urllib.parse import urlparse

import aiohttp
import discord
from discord.ext import commands

from poglink.config import MIN_POLLING_DELAY
from poglink.error import RatesFetchError, RatesProcessError
from poglink.models import RatesStatus

logger = logging.getLogger(__name__)

EMBED_IMAGE = "https://i.stack.imgur.com/Fzh0w.png"

# create cog class
class Rates(commands.Cog):
    DEFAULT_SERVER_INFO = {
        "smalltribes": {
            "short_name": "Smalltribes",
            "color": 0xA34C44,
        },
        "arkpocalypse": {
            "short_name": "Arkpocalypse",
            "color": 0xF8DE74,
        },
        "conquest": {
            "short_name": "Conquest/Classic",
            "color": 0x6DFF90,
        },
        None: {
            "short_name": "Official",
            "color": 0x63BCC3,
        },
    }

    def __init__(self, client):
        self.client = client

        self.webpage_urls = client.config.rates_urls
        self.channel_id = client.config.rates_channel_id
        self.polling_delay = client.config.polling_delay
        self.allowed_roles = client.config.allowed_roles
        self.data_dir = os.path.expanduser(client.config.data_dir)
        self.output_paths = [
            os.path.join(
                self.data_dir,
                f"last_{os.path.splitext(os.path.basename(urlparse(url).path))[0]}.json",
            )
            for url in client.config.rates_urls
        ]
        self.last_rates = [None for _ in self.webpage_urls]
        self.stable_rates = [None for _ in self.webpage_urls]
        self.consecutive_count = [0 for _ in self.webpage_urls]
        logger.debug(f"URLs: {self.webpage_urls}, Output paths: {self.output_paths}")
        # Create parent directory for persistent data if it doesn't exist yet
        if not os.path.exists(self.data_dir):
            logger.info(f"Data directory doesn't exist yet; creating: {self.data_dir}")
            os.makedirs(self.data_dir)

    @staticmethod
    async def get_current_rates(url):
        logger.debug(f"Requesting current server rates")
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    url,
                    headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
                ) as response:
                    response = await response.text()
                    rates = RatesStatus.from_raw(response)
                    logger.debug(f"Obtained server rates: {RatesStatus}")
                    return rates
            except ValueError as e:
                raise RatesProcessError(e) from e
            except Exception as e:
                raise RatesFetchError(e) from e

    async def send_embed(self, description, url):
        # generate embed
        logger.debug(f"Attempting to send embed. desc: {description}, url: {url}")
        try:
            match = re.match(
                "(?P<host>.*\/)?(?:(?P<platform>.*)\_(?P<game_mode>.*)\_)?dynamicconfig\.ini",
                os.path.basename(urlparse(url).path),
            )
            if match:
                server_match_dict = match.groupdict()
            else:
                server_match_dict = {}

            server_meta = self.DEFAULT_SERVER_INFO.get(
                server_match_dict.get("game_mode")
            )
            logger.debug(f"Server meta: {server_meta}")
            # TODO: Add ability to accept custom rates URL

        except Exception as e:
            logger.error(f"Rates url could not be processed: {url} {e}")
            return

        server_name = server_meta.get("short_name")

        embed = discord.Embed(
            description=description,
            title=f"ARK's {server_name} server rates have just been updated!",
            color=server_meta.get("color"),
        )
        embed.set_image(url=EMBED_IMAGE)

        # send embed
        channel = self.client.get_channel(self.channel_id)
        message = await channel.send(embed=embed)

        # if in announcement channel, publish message
        if message.channel.type == discord.ChannelType.news:
            logger.info("Announcement channel detected: Publishing message")
            await message.publish()

    async def compare_and_notify_all(self):
        for idx in range(len(self.webpage_urls)):
            url = self.webpage_urls[idx]

            # Fetch current rates from online
            try:
                current_rates = await self.get_current_rates(url)
            except RatesFetchError as e:
                logger.error(f"Could not retrieve rates from ARK Web API at {url}: {e}")
                continue
            except RatesProcessError as e:
                logger.error(f"Could not process rates URL {url}: {e}")
                continue
            except Exception as e:
                logger.error(e)
                continue

            # Only if last rates exist, get diff
            if self.last_rates[idx]:
                rates_diff = self.last_rates[idx].get_diff(current_rates)

                # If there is a difference, reset consecutive count; otherwise increment it
                if rates_diff.items:
                    self.consecutive_count[idx] = 1
                    self.last_rates[idx] = current_rates
                    logger.info(
                        f"Rates for {url} changed. Resetting consecutive rates count to 1"
                    )
                else:
                    self.consecutive_count[idx] += 1
                    logger.info(
                        f"Rates for {url} unchanged. Consecutive count = {self.consecutive_count[idx]}"
                    )

                if self.consecutive_count[idx] == 2:
                    logger.info(
                        f"New rates have become stable. Comparing against previous stable rates."
                    )
                    if self.stable_rates[idx]:
                        stable_diff = self.stable_rates[idx].get_diff(current_rates)
                        if stable_diff.items:
                            # generate and send embed
                            logger.info(
                                f"Rates at {url} changed since last stable value - sending embed"
                            )
                            embed_description = stable_diff.to_embed()
                            await self.send_embed(embed_description, url)
                    else:
                        logger.info(
                            f"No previous stable rates recorded at {url}. Updating new stable value, but no updates to publish."
                        )

                    self.stable_rates[idx] = current_rates
            else:
                self.consecutive_count[idx] = 1
                logger.info(f"No previous rates stored yet; skipping.")

            # Update last rates value for next iteration
            self.last_rates[idx] = current_rates

    # Events
    @commands.Cog.listener()
    async def on_ready(self):
        logger.info("Cog Ready: Rates")

        while True:
            await self.compare_and_notify_all()
            await asyncio.sleep(self.polling_delay)


# add cog to client
def setup(client):
    client.add_cog(Rates(client))



import aiohttp
import pytest

from poglink.cogs.bans import Bans


@pytest.fixture()
def bans_cog(sample_bot):
    bans_cog = Bans(sample_bot)
    return bans_cog


@pytest.mark.asyncio
async def test_bans_webpage_changed(bans_cog, configured_httpserver, last_bans):
    async with aiohttp.ClientSession() as session:
        async with session.get(
            configured_httpserver.url_for("/bansummary.txt"),
            headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
        ) as resp:
            assert resp.status == 200
            text = await resp.text()

    changed_status = await bans_cog.webpage_changed(text)
    assert text == last_bans
    assert changed_status == False

    async with aiohttp.ClientSession() as session:
        async with session.get(
            configured_httpserver.url_for("/bansummary-changed.txt"),
            headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
        ) as resp:
            assert resp.status == 200
            text = await resp.text()

    assert "PC Bans: 42069" in text

    changed_status = await bans_cog.webpage_changed(text)
    assert changed_status == True



import discord.ext.test
import pytest

from poglink.bot import ConfigurableBot


@pytest.fixture()
def sample_bot(bans_url_1, dynamic_rates_url, event_loop):
    # Define intents required to publish embeds
    intents = discord.Intents.default()
    intents.members = True

    bot = ConfigurableBot(
        ".",
        {
            "bans_url": bans_url_1,
            "rates_urls": [dynamic_rates_url],
            "data_dir": "tests/data",
        },
        loop=event_loop,  # Ensure pytest uses same event loop for async functions as bot
        intents=intents,
    )

    # Configure bot for testing
    discord.ext.test.configure(bot)

    # Override rates channel id with test channel id
    bot.config.rates_channel_id = bot.guilds[0].text_channels[0].id

    return bot


@pytest.fixture
def configured_httpserver(
    httpserver,
    sample_bansummary_1,
    sample_bansummary_2,
    sample_dynamicconfig_1,
    sample_dynamicconfig_2,
):
    httpserver.expect_request("/bansummary.txt").respond_with_data(
        sample_bansummary_1, content_type="text/plain"
    )
    httpserver.expect_request("/bansummary-changed.txt").respond_with_data(
        sample_bansummary_2, content_type="text/plain"
    )
    httpserver.expect_request("/dynamicconfig.ini").respond_with_data(
        sample_dynamicconfig_1, content_type="text/plain"
    )
    httpserver.expect_request("/dynamicconfig-changed.ini").respond_with_data(
        sample_dynamicconfig_2, content_type="text/plain"
    )

    return httpserver


@pytest.fixture
def rates_url_1(configured_httpserver):
    return configured_httpserver.url_for("/dynamicconfig.ini")


@pytest.fixture
def rates_url_2(configured_httpserver):
    return configured_httpserver.url_for("/dynamicconfig-changed.ini")


@pytest.fixture
def bans_url_1(configured_httpserver):
    return configured_httpserver.url_for("/bansummary.text")


@pytest.fixture
def bans_url_2(configured_httpserver):
    return configured_httpserver.url_for("/bansummary-changed.text")


@pytest.fixture
def sequential_handler(request):
    if hasattr(request, "param"):
        responses = request.param
    else:
        responses = []
        # Any test that requests this fixture should perform indirect parameterization to set the sequence of expected responses.

    # Create a generator out of sequence of respnoses
    resp_generator = (v for v in responses)

    # Handler returns next item in generator, and state is maintained until fixture is destroyed
    def my_handler(req):
        return next(resp_generator)

    return my_handler


@pytest.fixture
def sequential_httpserver(httpserver, sequential_handler):
    httpserver.expect_request(
        "/dynamic-dynamicconfig.ini", method="GET"
    ).respond_with_handler(sequential_handler)
    return httpserver


@pytest.fixture
def dynamic_rates_url(sequential_httpserver):
    return sequential_httpserver.url_for("/dynamic-dynamicconfig.ini")



import argparse
import os


def parse_list(raw):
    """Takes a string representation of a comma separated list and returns a list
    of elements with trimmed whitespace.

    Args:
        raw (str): Comma separated list as a string.

    Returns:
        List[str]: List of trimmed strings.
    """
    return [word.strip() for word in raw.split(",")]


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="""
        Bot for sending messages to Discord whenever ARK Web API changes.
        
        This bot requires some configuration parameters to be set. They can either
        be provided via a config file, CLI args, or environment variables. These
        values are set using the following order of priority:

            1. CLI Arguments
            2. Configuration file
            3. Environment Variables
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--data-dir",
        help="Directory in which to store persistent data and config file. Defaults to ~/.poglink",
    )
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        help="API token for ARK. Also can be set via BOT_TOKEN environment variable.",
    )
    parser.add_argument(
        "-p",
        "--polling-delay",
        type=float,
        help="Polling period in seconds. Also can be set via BOT_POLLING_PERIOD environment variable.",
    )
    parser.add_argument(
        "-a",
        "--allowed-roles",
        type=parse_list,
        help="Comma-separated list of allowed roles. Also can be set via BOT_ALLOWED_ROLES environment variable.",
    )
    parser.add_argument(
        "--rates-urls",
        type=parse_list,
        help="Comma-separated list of URLs for ARK rates. Also can be set via BOT_RATES_URLS environment variable.",
    )
    parser.add_argument(
        "--rates-channel-id",
        help="Discord 'rates' channel ID. Also can be set via BOT_RATES_CHANNEL_ID environment variable.",
    )
    # parser.add_argument(
    #     "--bans-url",
    #     help="URL for ARK bans. Also can be set via BOT_BANS_URL environment variable.",
    # ) # TODO: Reimplement when bans are enabled
    # parser.add_argument(
    #     "--bans-channel-id",
    #     help="Discord 'bans' channel ID. Also can be set via BOT_BANS_CHANNEL_ID environment variable.",
    # ) # TODO: Reimplement when bans are enabled
    parser.add_argument("--debug", action="store_true", help="Set log level to DEBUG.")

    return parser


def rotate_backups(full_path, max_copies=2):
    """Rotates backup files by appending sequential integer extensions.

    Example: If ``full_path`` is ``/my/path/to/file.txt``, then calling this function will
    rename / back up the file to ``/my/path/to/file.txt.1``. If another file is then written as
    ``file.txt`` and this function is called again, it will rename the previous backup to
    ``/my/path/to/file.txt.2``, and rename the new file to ``/my/path/to/file.txt.1``.

    The ``max_copies`` parameter controls to what limit this rolling backup is performed. Files
    in excess of this limit are deleted, in a first-in-first-out manner. The ``max_copies`` parameter
    considers the main file in addition to its backups. That is, if ``max_copies`` is set to 3, there
    will never be more than:

    - /my/path/to/file.txt
    - /my/path/to/file.txt.1
    - /my/path/to/file.txt.2

    Args:
        full_path (str): Full path to base file being rotated
        max_copies (int, optional): Maximum number of files to keep (backups + main). Defaults to 2.
    """
    full_path = os.path.expanduser(full_path)

    for i in reversed(range(max_copies)):
        if i == 0:
            old_filename = full_path
        else:
            old_filename = f"{full_path}.{i}"
        new_filename = f"{full_path}.{i+1}"

        if os.path.isfile(old_filename):
            if i == max_copies - 1:
                os.remove(old_filename)
                continue

            os.rename(old_filename, new_filename)
